/*
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992-1996 Takashi MANABE (manabe@papilio.tutics.tut.ac.jp)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY TAKASHI MANABE ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

#include	<config.h>

#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<unistd.h>
#include        <errno.h>

#include	<getcap.h>

#include	<defs.h>
#include	<version.h>
#include	<vc.h>

#define _USE_BSD
#include <sys/types.h>
#include <sys/resource.h>
#include <sys/wait.h>

static char *startupStr, *execProg;
static char *execProgArgs[256];

int	ConfigExecProg(const char *string, char *args[])
{
	int i=0;
	execProg = strdup(string);
	for (i = 0; i < 256; i++) {
	  if (args[i]) 
	      execProgArgs[i] = args[i];
	  else
	      break;
	}
	execProgArgs[i]=NULL;
	return SUCCESS;
}

#ifndef	MINI_KON

static int	ConfigStartup(const char *string)
{
	startupStr = strdup(string);
	return SUCCESS;
}

void	RunStartupCmd(void)
{
	char *p;
	int StartupPid;
	int stat;

	p = strtok(startupStr, "\n");
	while(p) {
	  /* fork handler */
	  if ((StartupPid = fork()) < 0) {
	    PerrorExit("fork");
	  }
	  if (StartupPid != 0) {
	    /* I'm parent. */

	    wait4(StartupPid, &stat, 0, 0);
	    if(WIFEXITED(stat) == 0)
	      PerrorExit("StartupCmd");

	  } else {

	    /* I'm child */
	    char *argv[4];

	    /* run as user, not effective user root */
	    setgid(getgid());
	    setuid(getuid());

	    argv[0] = "sh";
	    argv[1] = "-c";
	    argv[2] = p;
	    argv[3] = NULL;
	    execv("/bin/sh", argv);
	    _exit(0);
	  }
	  p = strtok(NULL, "\n");
	}
}

static bool startupMessage;

static int	ConfigMessage(const char *confstr)
{
	startupMessage = BoolConf(confstr);
	return SUCCESS;
}

void	ChildInit(void)
{
	DefineCap("StartupMessage", ConfigMessage, "On");
	DefineCap("StartUp", ConfigStartup, NULL);
}

#endif

void	ChildCleanup(void)
{
	free(startupStr);
}

void	ChildStart(FILE *errfp)
{
	char	*tail, *tcap;
	char	buff[256];
	extern char *TermName();

#ifndef	MINI_KON
	char	*shell;
        setgid(getgid());
        setuid(getuid());
	/* run as user, not effective user root */
#endif

#if defined(linux)
#ifdef	MINI_KON
	strcpy(buff, "TERM=linux");	/* ok - buff[256] */
#else
	strcpy(buff, "TERM=kon");	/* ok - buff[256] */
#endif
#elif defined(__FreeBSD__)
	sprintf(buff,"TERM=vt100");	/* ok - buff[256] */
#endif

	tcap = strdup(buff);
	putenv(tcap);

#ifndef	MINI_KON
	if (startupMessage)
	    printf("\rKON2 Kanji On Console " VERSION
		   " using VT number %c\r\n"
		   "%*s\r\n"
		   "%*s\r\n", *(TermName()+8),
		   dInfo.txmax,
		   "Copyright (C) "
		   "1993-2000  Takashi MANABE",
		   dInfo.txmax,
		   "1993, 1994 MAEDA Atusi   ");
#if defined(__FreeBSD__)
	printf("\rKON for FreeBSD-2.x ver0.01 Takashi OGURA\r\n");
#endif

	fflush(stdout);
#endif

	if (execProg)
	    execvp(execProg, execProgArgs);
	else {
	    if ((execProg = getenv("SHELL")) == NULL)
		execProg = "/bin/sh";
	    if ((tail = rindex(execProg, '/')) == NULL)
		tail = " sh";
	    snprintf(buff, sizeof(buff), "-%s", tail + 1);
	    execl(execProg, buff, 0);
	}
	fprintf(errfp, "KON> couldn't exec shell\r\n");
	fprintf(errfp, "%s: %s\r\n", execProg, strerror(errno));
	exit(EXIT_FAILURE);
}
